// io.cpp
//
//#undef BOCHS		// In ambiente Windows decommentare la definizione
#define BOCHS		//  giusta a seconda che si voglia generare uma 
			//  immagine funzionante (porta seriale e 
			//  riconoscimento hard disk) con l'emulatore Bochs 
			//  o con un PC reale.
			//  Se si dispone di make e` sufficiente $make BOCHS=no
			//  (vedere il README)

////////////////////////////////////////////////////////////////////////////////
//                                 COSTANTI                                   //
////////////////////////////////////////////////////////////////////////////////

// Livelli di privilegio dei processi
// Devono coincidere con i valori in sistema.cpp
//
const char LIV_UTENTE = 3, LIV_SISTEMA = 0;

// Interruzione hardware a priorita' minore 
//
const int IRQ_MAX = 15;

// Priorita' base dei processi esterni
// La priorita' di un processo esterno e' calcolata aggiungendo a questa
//  costante la priorita' dell' interruzione a cui e' associato (la priorita'
//  delle interruzioni e' inversamente proporzionale al loro tipo, quindi si
//  somma IRQ_MAX - TIPO alla priorita' di base).
// Deve coincidere con il valore in sistema.cpp
//
const int PRIO_ESTERN_BASE = 0x400;

// Identificatori dei processi esterni
// Devono coincidere con i valori in sistema.cpp
//
enum estern_id {
	tastiera,
	com1_in,
	com1_out,
	com2_in,
	com2_out,
	floppy,
	ata0,
	ata1,
	sb
};

typedef char* ind_b;		// indirizzo di un byte (una porta)
typedef short* ind_w;		// indirizzo di una parola
typedef int* ind_l;		// indirizzo di una parola lunga
typedef unsigned int ind_fisico;// locazione di memoria fisica (per il DMA)
typedef unsigned int size_t;

////////////////////////////////////////////////////////////////////////////////
//                        CHIAMATE DI SISTEMA USATE                           //
////////////////////////////////////////////////////////////////////////////////

extern "C" void terminate_p(void);
extern "C" void sem_ini(int &index_des_s, int val, bool &risu);
extern "C" void sem_wait(int sem);
extern "C" void sem_signal(int sem);

////////////////////////////////////////////////////////////////////////////////
//               INTERFACCIA OFFERTA DAL NUCLEO AL MODULO DI IO               //
////////////////////////////////////////////////////////////////////////////////

extern "C" void activate_pe(void f(int), int a, int prio, char liv,
	short &identifier, estern_id interf, bool &risu);

enum controllore { master=0, slave=1 };
extern "C" void nwfi(controllore c);

extern "C" bool verifica_area(void *area, unsigned int dim, bool write);
extern "C" void trasforma(ind_l vetti, ind_fisico &iff);
extern "C" void panic(const char *msg);
extern "C" int printk(const char *fmt, ...);
extern "C" void fill_gate(int gate, void (*f)(void), int tipo, int dpl);
extern "C" void reboot(void);
extern "C" bool reset_audio(int porta);

//
// Chiamate per l' IO da console
//

const int CON_BUF_SIZE = 0x0fa0;

struct con_status {
	unsigned char buffer[CON_BUF_SIZE];
	short x, y;
};

extern "C" void con_read(char &ch, bool &risu);
extern "C" void con_write(const char *vett, int quanti);

extern "C" void con_save(con_status *cs);
extern "C" void con_load(const con_status *cs);
extern "C" void con_update(con_status *cs, const char *vett, int quanti);
extern "C" void con_init(con_status *cs);

////////////////////////////////////////////////////////////////////////////////
//                      FUNZIONI GENERICHE DI SUPPORTO                        //
////////////////////////////////////////////////////////////////////////////////


// protezione dalla concorrenza con le interruzioni
//
inline void lock() { asm("cli"); }
inline void unlock() { asm("sti"); }

// ingresso di un byte da una porta di IO
extern "C" void inputb(ind_b reg, char &a);

// uscita di un byte su una porta di IO
extern "C" void outputb(char a, ind_b reg);

// ingresso di una word da una porta di IO
extern "C" void inputw(ind_b reg, short &a);

// uscita di una word su una porta di IO
extern "C" void outputw(short a, ind_b reg);

// ingresso di una stringa di word da un buffer di IO
extern "C" void inputbuffw(ind_w reg, short *a,short n);

// uscita di una stringa di word su un buffer di IO
extern "C" void outputbuffw(short *a, ind_w reg,short n);

void *memcpy(void *dest, const void *src, unsigned int n);
void *memset(void *dest, int c, size_t n);

////////////////////////////////////////////////////////////////////////////////
//                    GESTIONE DELLE INTERFACCE SERIALI                       //
////////////////////////////////////////////////////////////////////////////////

const int S = 2;

extern "C" void go_inputse(ind_b i_ctr);
extern "C" void halt_inputse(ind_b i_ctr);
extern "C" void go_outputse(ind_b i_ctr);
extern "C" void halt_outputse(ind_b i_ctr);

enum funz { input_n, input_ln, output_n, output_0 };

struct interfse_reg {
	ind_b iRBR;
	ind_b iTHR;
	ind_b iLSR;
	ind_b iIER;
	ind_b iIIR;
};

struct des_se {
	interfse_reg indreg;
	int mutex;
	int sincr;
	int cont;
	ind_b punt;
	funz funzione;
	char stato;
};

extern "C" des_se com[S];

void startse_in(des_se *p_des, char vetti[], int quanti, funz op)
{
	p_des->cont = quanti;
	p_des->punt = vetti;
	p_des->funzione = op;
	go_inputse(p_des->indreg.iIER);
}

extern "C" void c_readse_n(int serial, char vetti[], int quanti, char &errore)
{
	des_se *p_des;

	if(serial < 0 || serial >= S || !verifica_area(vetti, quanti, true) ||
			!verifica_area(&errore, sizeof(char), true))
		return;

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_in(p_des, vetti, quanti, input_n);
	sem_wait(p_des->sincr);
	errore = p_des->stato;
	sem_signal(p_des->mutex);
}

extern "C" void c_readse_ln(int serial, char vetti[], int &quanti, char &errore)
{
	des_se *p_des;

	if(serial < 0 || serial >= S ||
			!verifica_area(&quanti, sizeof(int), true) ||
			!verifica_area(vetti, quanti, true) ||
 			!verifica_area(&errore, sizeof(char), true))
		return;

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_in(p_des, vetti, 80, input_ln);
	sem_wait(p_des->sincr);
	quanti = p_des->cont;
	errore = p_des->stato;
	sem_signal(p_des->mutex);
}

void input_com(int h)
{
	char c, s;
	bool fine;
	des_se *p_des;

	p_des = &com[h];

	for(;;) {
		fine = false;
		halt_inputse(p_des->indreg.iIER);
		inputb(p_des->indreg.iRBR, c);
		inputb(p_des->indreg.iLSR, s);

		p_des->stato = s&0x1e;
		if(p_des->stato != 0)
			fine = true;
		else {
			if(p_des->funzione == input_n) {
				*p_des->punt = c;
				p_des->punt++;
				p_des->cont--;
				if(p_des->cont == 0)
					fine = true;
			} else if(p_des->funzione == input_ln)
				if(c == '\r' || c == '\n') {
					fine = true;
					p_des->cont = 80 - p_des->cont;
				} else {
					*p_des->punt = c;
					p_des->punt++;
					p_des->cont--;
					if(p_des->cont == 0) {
						fine = true;
						p_des->cont = 80;
					}
				}
		}

		if(fine == true) {
			*p_des->punt = 0;	// manca *
			sem_signal(p_des->sincr);
		} else
			go_inputse(p_des->indreg.iIER);

		nwfi(master); // sia com1 che com2 sono sul master
	}
}

// forza l' output del primo byte, sembra che l' interfaccia emulata da
//  Bochs presenti la richiesta di interruzione solo quando THR si svuota
//  dopo l' invio del primo byte
//
void kickse_out(des_se *p_des)
{
	char c = *p_des->punt;
	bool fine = false;

	p_des->punt++;

	if(p_des->funzione == output_n) {
		p_des->cont--;
		if(p_des->cont == 0) {
			halt_outputse(p_des->indreg.iIER);
			fine = true;
		}

		outputb(c, p_des->indreg.iTHR);
	} else if(p_des->funzione == output_0) {
		if(c == 0) {
			fine = true;
			halt_outputse(p_des->indreg.iIER);
		} else {
			p_des->cont++;
			outputb(c, p_des->indreg.iTHR);
		}
	}

	if(fine == true)
		sem_signal(p_des->sincr);
}

void startse_out(des_se *p_des, char vetto[], int quanti, funz op)
{
	p_des->cont = quanti;
	p_des->punt = vetto;
	p_des->funzione = op;
	go_outputse(p_des->indreg.iIER);
#ifdef BOCHS
	kickse_out(p_des);
#endif
}

extern "C" void c_writese_n(int serial, char vetto[], int quanti)
{
	des_se *p_des;

	if(serial < 0 || serial >= S || !verifica_area(vetto, quanti, false))
		return;

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_out(p_des, vetto, quanti, output_n);
	sem_wait(p_des->sincr);
	sem_signal(p_des->mutex);
}

extern "C" void c_writese_0(int serial, char vetto[], int &quanti)
{
	des_se *p_des;

	if(serial < 0 || serial >= S ||
			!verifica_area(&quanti, sizeof(int), true) ||
			!verifica_area(vetto, quanti, false))
		return;

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_out(p_des, vetto, 0, output_0);
	sem_wait(p_des->sincr);
	quanti = p_des->cont;
	sem_signal(p_des->mutex);
}

void output_com(int h)
{
	char c;
	bool fine;
	des_se *p_des;

	p_des = &com[h];	// como

	for(;;) {
		fine = false;
		if(p_des->funzione == output_n) {
			p_des->cont--;
			if(p_des->cont == 0) {
				fine = true;
				halt_outputse(p_des->indreg.iIER);
			}

			c = *p_des->punt;
			outputb(c, p_des->indreg.iTHR);
			p_des->punt++;
		} else if(p_des->funzione == output_0) {
			c = *p_des->punt;
			if(c == 0) {
				fine = true;
				halt_outputse(p_des->indreg.iIER);
			} else {
				outputb(c, p_des->indreg.iTHR);
				p_des->cont++;
				p_des->punt++;
			}
		}

		if(fine == true)
			sem_signal(p_des->sincr);

		nwfi(master); // sia com1 che com2 sono sul master
	}
}

extern "C" void com_setup(void);

const estern_id com_id[2][2] = {
	{ com1_in, com1_out },
	{ com2_in, com2_out }
};

// Interruzione hardware della seconda interfaccia seriale
//
const int COM2_IRQ = 4;

void com_init()
{
	des_se *p_des;
	bool r1, r2;
	short id;
	int i, com_base_prio = PRIO_ESTERN_BASE + IRQ_MAX - COM2_IRQ;

	com_setup();

	for(i = 0; i < S; ++i) {
		p_des = &com[i];

		sem_ini(p_des->mutex, 1, r1);
		sem_ini(p_des->sincr, 0, r2);
		if(!r1 || !r2)
			panic("Impossibile allocare semaforo per l' IO");

		activate_pe(input_com, i, com_base_prio - i,
			LIV_SISTEMA, id, com_id[i][0], r1);
		activate_pe(output_com, i, com_base_prio - i,
			LIV_SISTEMA, id, com_id[i][1], r2);
		if(!r1 || !r2)
			panic("Impossibile creare processo esterno di uscita");
	}
}

////////////////////////////////////////////////////////////////////////////////
//                         GESTIONE DELLA TASTIERA                            //
////////////////////////////////////////////////////////////////////////////////

#define KBD_MOD_NR	2		// numero di modificatori usati
#define KBD_COD_NR	127		// numero di codici

// modificatori
enum kbd_mod {
	KBD_MOD_NONE = 0,
	KBD_MOD_SHIFT = 1,
	KBD_MOD_ALT = 2,
	KBD_MOD_ALTGR = 3,
	KBD_MOD_ALTSH = 4,
	KBD_MOD_CTRL = 5
};

#define KBD_SHIFT	(KBD_MOD_SHIFT << 8)
#define KBD_ALT		(KBD_MOD_ALT << 8)
#define KBD_ALTGR	(KBD_MOD_ALTGR << 8)
#define KBD_ALTSH	(KBD_MOD_ALTSH << 8)
#define KBD_CTRL	(KBD_MOD_CTRL << 8)

#define KBD_MOD_MASK	0x0f00

#define KBD_IS_MOD(c)	((c&KBD_MOD_MASK) != 0)
#define KBD_MOD(c)	((c&KBD_MOD_MASK) >> 8)

// tasti funzione
#define KBD_F1		0x1000
#define KBD_F2		0x2000
#define KBD_F3		0x3000
#define KBD_F4		0x4000
#define KBD_F5		0x5000
#define KBD_F6		0x6000
#define KBD_F7		0x7000
#define KBD_F8		0x8000
#define KBD_F9		0x9000
#define KBD_F10		0xa000
#define KBD_F11		0xb000
#define KBD_F12		0xc000

#define KBD_FNUM(code)	(((code) - KBD_F1) >> 12)

#define KBD_CANC	0xd000

// tabella di corrispondenza tra codici e caratteri ascii (tastiera italiana)
//
static unsigned short kbd_keymap[KBD_MOD_NR][KBD_COD_NR] = {
	{ 0, 27, '1', '2', '3', '4', '5', '6', '7', '8',
	'9', '0', '\'', 141, '\b', '\t', 'q', 'w', 'e', 'r',
	't', 'y', 'u', 'i', 'o', 'p', 138, '+', '\n', KBD_CTRL,
	'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 149,
	133, '\\', KBD_SHIFT, 151, 'z', 'x', 'c', 'v', 'b', 'n',
	'm', ',', '.', '-', KBD_SHIFT, '*', KBD_ALT, ' ', 0, KBD_F1,
	KBD_F2, KBD_F3, KBD_F4, KBD_F5, KBD_F6, KBD_F7, KBD_F8, KBD_F9,
		KBD_F10, 0,
	0, 0, 0, 0, '-', 0, 0, 0, '+', 0,
	0, 0, 0, KBD_CANC, '\n', 0, 0, KBD_F11, KBD_F12, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0 },

	{ 0, 27, '!', '"', 156, '$', '%', '&', '/', '(',
	')', '=', '?', '^', 0, 0, 'Q', 'W', 'E', 'R',
	'T', 'Y', 'U', 'I', 'O', 'P', 130, '*', 0, KBD_CTRL,
	'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 135,
	248, '|', KBD_SHIFT, 21, 'Z', 'X', 'C', 'V', 'B', 'N',
	'M', ';', ':', '_', KBD_SHIFT, '*', KBD_ALT, ' ', 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, '7', '8', '9', '-', '4', '5', '6', '+', '1',
	'2', '3', '0', '.', 0, 0, '>', 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0 }
};

// dimensione della coda di una tastiera
const int KBD_QUEUE_SIZE = 512;

// maschere per il campo flags
enum {
	KBD_FLAG_NONE = 0,
	KBD_FLAG_SHIFT = 0x100,
	KBD_FLAG_ALT = 0x200,
	KBD_FLAG_CTRL = 0x400
};

struct des_term;

// tastiera virtuale
struct kbd_t {
	char queue[KBD_QUEUE_SIZE];
	int in, out;
	des_term *term;
	int flags;
	int nl_cnt;
};

// flag in corrispondenza dei modificatori
const int kbd_modflag_tab[6] = {
	KBD_FLAG_NONE,
	KBD_FLAG_SHIFT,
	KBD_FLAG_ALT,
	KBD_FLAG_ALT,
	KBD_FLAG_NONE,
	KBD_FLAG_CTRL
};

// tastiera virtuale attiva
static kbd_t *kbd_att = 0;

// inizializzazione di una tastiera virtuale
inline void kbd_init(kbd_t *kbd, des_term *term)
{
	memset(kbd, 0, sizeof(kbd_t));
	kbd->term = term;
	if(!kbd_att)
		kbd_att = kbd;
}

// vera se e' stato premuto il ritorno a capo
inline bool KBD_NL_READ(kbd_t *kbd)
{
	return kbd->nl_cnt > 0;
}

// indice nella keymap in base a flags
inline int KBD_SC_IDX(int flags)
{
	return (flags)&KBD_FLAG_SHIFT ?
		KBD_MOD_SHIFT: KBD_MOD_NONE;
}

// inserimento del codice nella coda (di caratteri) della tastiera attiva
//
static void kbd_put(unsigned short code)
{
	char ch;

	if(!kbd_att)
 		return;

	ch = (char)(code&0xff);
	switch(ch) {
		case '\b':
			if(kbd_att->in != kbd_att->out &&
					kbd_att->queue[kbd_att->in > 0 ?
     					kbd_att->in - 1: KBD_QUEUE_SIZE - 1] != '\n') {
				--kbd_att->in;
				kbd_att->in = kbd_att->in >= 0 ?
					kbd_att->in: KBD_QUEUE_SIZE - 1;
			}

			break;
		default:
			if(!ch)
				return;

			kbd_att->queue[kbd_att->in] = ch;
			if(ch == '\n')
				++kbd_att->nl_cnt;
			kbd_att->in = (kbd_att->in + 1) % KBD_QUEUE_SIZE;
			if(kbd_att->in == kbd_att->out)
				kbd_att->out = (kbd_att->out + 1) %
					KBD_QUEUE_SIZE;
	}
}

// lettura di un carattere dalla coda di kbd
//
static int kbd_get(kbd_t *kbd)
{
	int rv;

	if(kbd->in == kbd->out)
		return -1;

	rv = kbd->queue[kbd->out];
	if(rv == '\n')
		--kbd->nl_cnt;
	kbd->out = (kbd->out + 1) % KBD_QUEUE_SIZE;

	return rv;
}

bool term_newchar(des_term *term, unsigned short code);

// processo esterno per la gestione dell' ingresso da tastiera
//
void tast_in(int h)
{
	char ch;
	unsigned short code;
	bool r;

	for(;;) {
		con_read(ch, r);

		if(r && kbd_att) {
			code = kbd_keymap[KBD_SC_IDX(kbd_att->flags)][ch&0x7f];

			if(!(ch&0x80) && KBD_IS_MOD(code))
				kbd_att->flags |= kbd_modflag_tab[KBD_MOD(code)];
			else if((ch&0x80) && KBD_IS_MOD(code))
				kbd_att->flags &= ~kbd_modflag_tab[KBD_MOD(code)];
			else if(!(ch&0x80)) {
				if(term_newchar(kbd_att->term, code))
					kbd_put(code);
			}
		}

		nwfi(master);
	}
}

extern "C" void abilita_tastiera(void);

// Interruzione hardware della tastiera
//
const int KBD_IRQ = 1;

// inizializzazione
//
void kbd_init()
{
	short id;
	bool r;

	activate_pe(tast_in, 0, PRIO_ESTERN_BASE + IRQ_MAX - KBD_IRQ,
		LIV_SISTEMA, id, tastiera, r);
	if(!r)
		panic("Impossibile creare processo esterno della tastiera");

	// l' inizializzazione avviene ad interruzioni disabilitate, quindi
	//  non ci saranno richieste di interruzione accolte fino all' uscita
	//  da io_init
	//
	abilita_tastiera();
}

// lettura di n caratteri da kbd
//  (deve essere chiamata ad interruzioni disabilitate)
//
int kbd_read(kbd_t *kbd, char *buf, int n)
{
	int i = 0, ch;

	while(i < n && (ch = kbd_get(kbd)) != -1) {
		buf[i++] = (char)ch;
		if(ch == '\n')
			break;
	}

	return i;
}

// numero di caratteri disponibili nella coda di kbd
//
int kbd_avail(kbd_t *kbd)
{
	if(kbd->in >= kbd->out)
		return kbd->in - kbd->out;

	return KBD_QUEUE_SIZE - kbd->out + kbd->in;
}

// rende kbd la tastiera virtuale attiva
//
int kbd_activate(kbd_t *kbd)
{
	kbd_att->flags = KBD_FLAG_NONE;
	kbd_att = kbd;

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//                             I/O DA TERMINALE                               //
////////////////////////////////////////////////////////////////////////////////

// descrittore di terminale virtuale
//
struct des_term {
	int mutex;	// protezione da utilizzo concorrente
	int sincr;	// sincronizzazione per la lettura da tastiera
	bool waiting;	// un processo attende su sincr

	kbd_t kbd;	// tastiera virtuale
	con_status cstat; // stato della console per il video virtuale
};

const int N_TERM = 12;
int term_att = 0;

des_term term_virt[N_TERM];

void attiva_term(int t)
{
	con_save(&term_virt[term_att].cstat);
	term_att = t;
	kbd_activate(&term_virt[term_att].kbd);
	con_load(&term_virt[term_att].cstat);
}

bool term_newchar(des_term *term, unsigned short code)
{
	char ch;

	if(code >= KBD_F1 && code <= KBD_F12) {
		attiva_term(KBD_FNUM(code));
		return false;
	}

	if(code == KBD_CANC && (term->kbd.flags&KBD_FLAG_CTRL) &&
			(term->kbd.flags&KBD_FLAG_ALT))
		reboot();

	if(code == '\b' && kbd_avail(&term->kbd) == 0)
		return false;

	if(code < 0xff) {
		ch = code&0xff;
		con_write(&ch, 1);
	}

	if(term->waiting && code == '\n') {
		term->waiting = false;
		sem_signal(term->sincr);
	}

	return true;
}

// NOTA: non richiede che vetti sia nello spazio condiviso, dato che l' handler
//  dell' interruzione scrive in un buffer interno
extern "C" void c_term_read_n(int term, char vetti[], int &quanti)
{
	des_term *p_des;
	int disp, letti;

	if(term < 0 || term >= N_TERM ||
			!verifica_area(&quanti, sizeof(int), true) ||
			!verifica_area(vetti, quanti, true))
		return;

	p_des = &term_virt[term];
	sem_wait(p_des->mutex);
	quanti = quanti > KBD_QUEUE_SIZE ? KBD_QUEUE_SIZE: quanti;

	lock();
	disp = kbd_avail(&p_des->kbd);

	if(disp >= quanti - 1 || KBD_NL_READ(&p_des->kbd)) {
		// sono gia' disponibili abbastanza caratteri, o
  		//  e' stato premuto invio
		letti = kbd_read(&p_des->kbd, vetti, quanti - 1);
	} else {
		// e' necessario bloccare
		p_des->waiting = true;
		unlock();

		sem_wait(p_des->sincr);

		lock();
		disp = kbd_avail(&p_des->kbd);
		letti = kbd_read(&p_des->kbd, vetti, quanti - 1);
	}

	unlock();

	vetti[letti] = 0;
	quanti = letti + 1;

	sem_signal(p_des->mutex);
}

// NOTA: non richiede che vetti sia nello spazio condiviso, dato che l' IO non
//  e' ad interruzione di programma
extern "C" void c_term_write_n(int term, char vetti[], int quanti)
{
	des_term *p_des;

	if(term < 0 || term >= N_TERM || !verifica_area(vetti, quanti, false))
		return;

	p_des = &term_virt[term];
	sem_wait(p_des->mutex);

	lock();
	if(term_att == term)
		con_write(vetti, quanti);
	else
		con_update(&term_virt[term].cstat, vetti, quanti);
	unlock();

	sem_signal(p_des->mutex);
}

void term_init(void)
{
	des_term *p_des;
	bool r1, r2;

	kbd_init();

	for(int i = 0; i < N_TERM; ++i) {
		p_des = &term_virt[i];

		sem_ini(p_des->mutex, 1, r1);
		sem_ini(p_des->sincr, 0, r2);
		if(!r1 || !r2)
			panic("Impossibile inizializzare i semafori per l' I/O");
		p_des->waiting = false;

		kbd_init(&p_des->kbd, p_des);
		con_init(&p_des->cstat);
   	}
}

////////////////////////////////////////////////////////////////////////////////
//            		GESTIONE DEL MECCANISMO DI DMA               	      //
////////////////////////////////////////////////////////////////////////////////

#define READ_BLOCK 0x44		// 0b01000100 MODE = single(01),noautoinit(0),
				//  addr.inc(0),lettura(01),canale(xx)
#define WRITE_BLOCK 0x48	// 0b01001000 MODE = single(01),noautoinit(0),
				//  addr.inc(0),scrittura(10),canale(xx)
#define READ_BLOCKAI 0x54	// 0b01010100 MODE = single(01),noautoinit(0),
				//  addr.inc(0),lettura(01),canale(xx)
#define WRITE_BLOCKAI 0x58	// 0b01011000 MODE = single(01),noautoinit(0),
				//  addr.inc(0),scrittura(10),canale(xx)

const int D=2;

struct dma_reg {		// Registri del Controllore DMA comuni a tutti
	ind_b iCMD_STS;	
	ind_b iMODER;	
	ind_b iREQREG;	
	ind_b iMASK_REG;
	ind_b iMASK;	
	ind_b iMASTERRS;
	ind_b iMASKCLR;
};

struct dmachan_reg {		// Registri del Controllore DMA specifici per
	ind_b iPAGE;		//  ogni canale
	ind_b iMAR;
	ind_b iTCR;
	ind_b iCLEAR_FF;
};

extern "C" const unsigned char dma_chan_ind[8][4] =  {{0x00,0x01,0x87,0X0C},{0x02,0x03,0x83,0x0C},{0x04,0x05,0x81,0x0C},{0x06,0x07,0x82,0x0C},{0xc0,0xc2,0x8b,0xd8},{0xc4,0xc6,0x8b,0xd8},{0xc8,0xca,0x89,0xd8},{0xcc,0xce,0x8a,0xd8}};

extern "C" dma_reg cdma[D];	// Controllori DMA 8237 master(4-7) e slave (0-3)

extern "C" void dma_init();

// Trasferimento dati in DMA
//

extern "C" void setup_addr_dma(ind_b p,ind_b m,ind_fisico i_f);

extern "C" void setup_cont_dma(ind_b c,int c);

extern "C" void go_moder_in(unsigned char m,ind_b c,ind_b msk,short int ch);

extern "C" void go_moder_out(unsigned char m,ind_b c,ind_b msk,short int ch);

extern "C" void halt_moder(ind_b c,ind_b msk,short int ch);
	
inline void dma_status(ind_b reg,char &s) {
	inputb(reg,s);
}

////////////////////////////////////////////////////////////////////////////////
//               		GESTIONE DEL FLOPPY DISK               	      //
////////////////////////////////////////////////////////////////////////////////

// Codici degli errori comuni alle unita di memorizzazione di massa

#define D_ERR_NONE 0x00
#define D_ERR_BOUNDS 0xFF
#define D_ERR_PRESENCE 0xFE
// Specifici per il floppy
#define FD_ERR_NORECAL 0xFD
#define FD_ERR_NOSEEK 0xFC
#define FD_ERR_NOADDR 0xFB
#define FD_ERR_ADDRDMA 0xFA
#define FD_ERR_NOTX 0xF9

#define BLK_SIZE 512			// Dimensione del blocco (HD e FD)
#define H_BLK_SIZE (BLK_SIZE / 2)	// Meta' dimensione blocco (short int)

#define D_TIMEOUT 0x7FFF		// Timeout per l'accesso ai dischi
#define FD_RECAL_TIMEOUT 0xFF		// Timeout per il recalibrate del floppy

#define FD_STS_SIZE 7
#define FD_N_BLK 2880

// Geometria del floppy (fissa, floppy 1.44MB)
#define FD_TEST 2
#define FD_TR 80
#define FD_SETT_T 18
#define FD_MAX_BLK 128

// Operazione di lettura o scrittura
#define FDC_WRITE false
#define FDC_READ true
#define FD_RETRY 3		// Numero di tentadivi del driver del floppy

enum fd_cmd {			// Comandi per il controllore del floppy
	SPECIFY=0x03,
	RECALIBRATE=0x07,
	SEEK=0x0F,
	SENSEI=0x08,
	READ=0xE6,
	WRITE=0xC5,
	FORMAT=0x4D
};
	
struct fdcontr_reg {		// Registri del contrlollore del floppy
	ind_b iDOR;
	union {
		ind_b iM_STS;
		ind_b iD_RATE;
	}; //iM_STS_D_RATE
	ind_b iDATA;
	union {
		ind_b iDIR;
		ind_b iCCR;
	}; //iDIR_CCR
};

struct fddrive_data {		// Informazioni specifiche di un floppy drive
	bool motore;
	bool presente;
	char errore;
	int traccia;
};

struct des_fd {			// Descrittore di operazione di I/O su floppy
	fdcontr_reg indreg;
	dmachan_reg dmareg;
	fddrive_data drive[2];		// Gestiti fino a 2 drive
	fd_cmd curr_cmd;
	int curr_drv;
	int mutex;
	int sincr;
	int dma;
};

extern "C" des_fd fd;		// gestito un solo controllore

//Funzioni di utilita' per la gestione del disco
//

extern "C" int test_fd();

extern "C" void abilita_fd();

inline void fdc_delay() {
	char c;
	inputb((char*)0x80,c);
}

// Converte un blocco in testina-traccia-settore
//
void blk_tts(int primo, int &test, int &tr, int &sett) {
		//inversa di: primo=(2*FD_TR+FD_TEST)*FD_SETT_T+sett-1)
		test=(primo%(FD_SETT_T*FD_TEST))/(FD_SETT_T);
		tr=primo/(FD_SETT_T*FD_TEST);
		sett=(primo%FD_SETT_T)+1;
}

// Lettura dei blocchi non ancora trasferiti dal dma
// 
inline void get_quanti(des_fd *p_des,unsigned char &q) {
	char contL,contH;
	outputb(0x00,p_des->dmareg.iCLEAR_FF);
	inputb(p_des->dmareg.iTCR,contL);
	inputb(p_des->dmareg.iTCR,contH);
	q=(contH<<8)+(contL+1);
}

// Uscita di un byte verso il controllore floppy
//  la procedura richiede una temporizzazione precisa!
//
void outputb_fd(char a) {
	char stato;
	des_fd *p_des=&fd;
	short timeout=D_TIMEOUT;
	do {
		fdc_delay();			// Ritardo!
		inputb(p_des->indreg.iM_STS,stato);
		timeout--;
		if (timeout<0)
			panic("OUTPUT: Floppy disk timeout!\n");
	} while ((stato&0xC0)!=0x80);
	outputb(a,p_des->indreg.iDATA);
}

// Ingresso di un byte dal controllore floppy
//  la procedura richiede una temporizzazione precisa!
//
void inputb_fd(char &a) {
	char stato;
	des_fd *p_des=&fd;
	
	short timeout=D_TIMEOUT;
	do {
		fdc_delay();			// Ritardo!
		inputb(p_des->indreg.iM_STS,stato);
		timeout--;
		if (timeout<0)
			panic("INPUT: Floppy disk timeout!\n");
	} while ((stato&0xD0)!=0xD0);
	inputb(p_des->indreg.iDATA,a);
}

extern "C" void motore_on(ind_b c,bool &m,int d);

extern "C" void motore_off(ind_b c,bool &m);

// Invio di un comando (ultimo byte) al controllore e aggiornamento del
//  descrittore dell'operazione
//  
inline void invia_cmd_fd(des_fd *p_des,fd_cmd com) {
	outputb_fd(com);
	p_des->curr_cmd=com;
}

// Lettura dello stato del controllore floppy; da compiere in risposta
//  ad alcuni comandi
//  
void fdc_status(char buff[FD_STS_SIZE]) {
	char stato;
	des_fd *p_des=&fd;
	
	for (int i=0;i<FD_STS_SIZE;i++) {
		inputb(p_des->indreg.iM_STS,stato);
		if (!(stato&&0x10))
				break;
		inputb_fd(buff[i]);
	}
}

// Reset del piedino di interrupt del floppy e lettura dello stato del
//  dispositivo; da compiere in risposta ad alcuni comandi
//  
char fdc_sensei(int drv) {
	char stato;
	char curr_tr;
	des_fd *p_des=&fd;
	
	invia_cmd_fd(p_des,SENSEI);
	inputb_fd(stato);
	inputb_fd(curr_tr);
	if (drv!=-1)
		p_des->drive[drv].traccia=curr_tr;
	return stato;
}

// Riporta la testina sulla traccia 0; da compiere dopo il verificarsi di un
//  qualunque errore
//  
bool fdc_recalibrate(int drv,int test) {	// Ha un valore di ritorno 
	char stato;				//  potendo a sua volta fallire
	int timeout=0xFF;
	des_fd *p_des=&fd;
	
	if (!p_des->drive[0].motore)
		motore_on(p_des->indreg.iDOR,p_des->drive[drv].motore,drv);
	do {
		p_des->curr_drv=drv;
		invia_cmd_fd(p_des,RECALIBRATE);// quando e' possibile il floppy
		outputb_fd((test<<2)|drv);	//  deve funzionare ad
		sem_wait(p_des->sincr);		//  interruzione
		stato=p_des->drive[drv].errore;
	} while ((timeout--)>0 && (stato&0x30)==0x30);
	if ((stato&0xF8)!=0x20) {
		motore_off(p_des->indreg.iDOR,p_des->drive[drv].motore);
		p_des->drive[drv].errore=FD_ERR_NORECAL;
		return false;
	}
	p_des->drive[drv].errore=D_ERR_NONE;
	return true;
}

// Posizionamento della testina di lettura/scrittura; in molti floppy controller
//  non e' automatico come per gli hard disk
//  
bool fdc_seek(int drv,int test,int tr) {	// Prevede il ritorno anziche
	char stato;				//  bloccare il sistema
	des_fd *p_des=&fd;
	
	if (tr!=(p_des->drive[drv].traccia)) {
		p_des->curr_drv=drv;
		invia_cmd_fd(p_des,SEEK);
		outputb_fd((test<<2)|drv);
		outputb_fd(tr);
		sem_wait(p_des->sincr);		// Attesa dell'interruzione
		stato=p_des->drive[drv].errore;
		if ((stato&0xF8)!=0x20 || (p_des->drive[drv].traccia!=tr)) {
			motore_off(p_des->indreg.iDOR,p_des->drive[drv].motore);
			p_des->drive[drv].errore=FD_ERR_NOSEEK;
			return false;
		}
	}
	p_des->drive[drv].errore=D_ERR_NONE;
	return true;
}

// Impartisce un comando di lettura o scrittura al controllore; la funzione
//  e' unica essendo la struttura (peraltro complessa) dei due comandi molto
//  simile, almeno fino all'ultimo byte da inviare
//  
void fdc_read_write(int drv,bool read,int test,int tr,int sett) {
	char comando[8];
	des_fd *p_des=&fd;
	
	// Invio del comando di lettura o di scrittura
	if (read)
		invia_cmd_fd(p_des,READ);
	else
		invia_cmd_fd(p_des,WRITE);
	comando[0]=(test<<2)|drv;	// Scelta del drive
	comando[1]=tr;			// Indirizzamento
	comando[2]=test;
	comando[3]=sett;
	comando[4]=2;			// 512 byte per settore
	comando[5]=18;			// Settori per traccia -> buffer FIFO
	comando[6]=0x1B;		// disco da 1.44 MB
	comando[7]=0xFF;		// DTL non utilizzato
	p_des->curr_drv=drv;
	for (int i=0;i<8;i++)
		outputb_fd(comando[i]);
}

// Avvio delle operazioni di lettura
//  (NB: puo' anche questa fallire)
//  
int startfd_in(des_fd *p_des,dma_reg *d_des,short drv,ind_w vetti,unsigned int primo,unsigned char &quanti) {
	char stato;
	ind_fisico iff;
	int test=0,tr=0,sett=1;			
	blk_tts(primo,test,tr,sett);		// Traduzione indirizz
	//Attivazione del motorino
	motore_on(p_des->indreg.iDOR,p_des->drive[drv].motore,drv);	
	outputb(0x00,p_des->indreg.iCCR);	// DataRate 500Kb/sec
	if (!fdc_seek(drv,test,tr)) {		// La seek puo' fallire
		//sem_signal(p_des->sincr);	//  (processo esterno)
		return -1;
	}

	trasforma((ind_l)vetti,iff);		// Ottenimento dell'indirizzo
						//  fisico (DMA!!)
	if (iff==0) {
		p_des->drive[drv].errore=FD_ERR_ADDRDMA;
		return -1;
	}

	// Inizializzazione del controllore DMA (indirizzi e contatore)
	setup_addr_dma(p_des->dmareg.iPAGE,p_des->dmareg.iMAR,iff);
	setup_cont_dma(p_des->dmareg.iTCR,BLK_SIZE*quanti);
	go_moder_in(READ_BLOCK,d_des->iMODER,d_des->iMASK,2);	// Programmazione

	fdc_read_write(drv,FDC_READ,test,tr,sett);		// Comando di lettura

	return test;		// ritorna la testina, ad uso della primitiva
}

// Avvio delle operazioni di scrittura
//  (NB: anche questa puo' fallire)
//  
int startfd_out(des_fd *p_des,dma_reg *d_des,short drv,ind_w vetti,unsigned int primo,unsigned char &quanti) {
	char stato;
	ind_fisico iff;
	int test=0,tr=0,sett=1;
	
	blk_tts(primo,test,tr,sett);		// Traduzione indirizzo
	//Attivazione del motorino
	motore_on(p_des->indreg.iDOR,p_des->drive[drv].motore,drv);
	outputb(0x00,p_des->indreg.iCCR);	// DataRate 500Kb/sec
	if (!fdc_seek(drv,test,tr)) {		// La seek puo' fallire
		//sem_signal(p_des->sincr);	//  (processo esterno)
		return -1;
	}
	
	trasforma((ind_l)vetti,iff);		// Ottenimento dell'indirizzo
						//  fisico (DMA!)
	if (iff==0) {
		p_des->drive[drv].errore=FD_ERR_ADDRDMA;
		return -1;
	}

	// Inizializzazione del controllore DMA (indirizzi e contatore)
	setup_addr_dma(p_des->dmareg.iPAGE,p_des->dmareg.iMAR,iff);
	setup_cont_dma(p_des->dmareg.iTCR,BLK_SIZE*quanti);
	go_moder_out(WRITE_BLOCK,d_des->iMODER,d_des->iMASK,2);	// Programmazione
	
	fdc_read_write(drv,FDC_WRITE,test,tr,sett);	// Comando di scrittura

	return test;		// ritorna la testina, ad uso della primitiva

}

// Primitiva di lettura dal floppy: legge quanti blocchi (max 128)
//  a partire da primo e li deposita in vetti (attraverso l'ind. fisico);
//  legge dal drive drv e riporta un fallimento in errore
//  
extern "C" void c_readfd_n(short drv,short vetti[],unsigned int primo,unsigned char &quanti,char &errore)
{
	des_fd *p_des;
	dma_reg *d_des;

	p_des=&fd;
	d_des=&cdma[0];		// Floppy = Controllore DMA Slave

	// Controllo di protezione
	if (drv<0 || drv>=2)
		return;
	if (!verifica_area(&quanti,sizeof(int),true) ||
	    !verifica_area(vetti,quanti,true) ||
	    !verifica_area(&errore,sizeof(char),true))
		return;

	// Controllo sulla selezione di un drive presente
	if (!p_des->drive[drv].presente) {
		errore=D_ERR_PRESENCE;
		return;
	}
	
	// Controllo sull'indirizzamento
	if (primo+quanti>FD_N_BLK || quanti>FD_MAX_BLK) {
		errore=D_ERR_BOUNDS;
		return;
	}

	sem_wait(p_des->mutex);

	for (int i=0;i<=FD_RETRY;i++) {		// Tenta la lettura piu' volte
	p_des->drive[drv].errore=D_ERR_NONE;	// Reset errori precedenti
	// Avvio operazioni di lettura (la chiamata puo' fallire!)
	int test=startfd_in(p_des,d_des,drv,vetti,primo,quanti);
	if (test!=-1) {				// Se qualcosa e' andato storto
		sem_wait(p_des->sincr);		//  non ci blocchiamo
		if (p_des->drive[drv].errore==D_ERR_NONE)
			break;
		else
			if (i!=FD_RETRY)
				fdc_recalibrate(drv,test); // prima di riprovare
		}					
	}

	get_quanti(p_des,quanti);
	errore=p_des->drive[drv].errore;
	sem_signal(p_des->mutex);

}

// Primitiva di scrittura su floppy: scrive quanti blocchi (max 128)
//  a partire da primo prelevandoli da vetti (attraverso l'ind. fisico);
//  scrive sul drive drv e riporta un fallimento in errore
//  
extern "C" void c_writefd_n(short drv,ind_w vetti,unsigned int primo,unsigned char &quanti,char &errore)
{
	des_fd *p_des;
	dma_reg *d_des;

	p_des=&fd;
	d_des=&cdma[0];		// Floppy = Controllore DMA Slave

	// Controllo di protezione
	if (drv<0 || drv>=2)
		return;
	if (!verifica_area(&quanti,sizeof(int),true) ||
	    !verifica_area(vetti,quanti,true) ||
	    !verifica_area(&errore,sizeof(char),true))
		return;

	//Controllo sulla selezione di un drive presente
	if (!p_des->drive[drv].presente) {
		errore=D_ERR_PRESENCE;
		return;
	}
	
	//Controllo sull'indirizzamento
	if (primo+quanti>FD_N_BLK || quanti>FD_MAX_BLK) {
		errore=D_ERR_BOUNDS;
		return;
	}

	sem_wait(p_des->mutex);
	
	for (int i=0;i<=FD_RETRY;i++) {		// Tenta la scrittura piu' volte
	p_des->drive[drv].errore=D_ERR_NONE;	// reset errori precedenti
	// Avvio operazioni di lettura (la chiamata puo' fallire!)
	int test=startfd_out(p_des,d_des,drv,vetti,primo,quanti);
	if (test!=-1) {				// Se qualcosa e' andato storto
		sem_wait(p_des->sincr);		//  non ci blocchiamo
		if (p_des->drive[drv].errore==D_ERR_NONE)
			break;
		else
			if (i!=FD_RETRY)
				fdc_recalibrate(drv,test); // prima di riprovare
		}
	}
	
	get_quanti(p_des,quanti);
	errore=p_des->drive[drv].errore;
	sem_signal(p_des->mutex);
}

// Processo esterno per il floppy disk (in DMA)
// 
void dmaest_fd(int h) {
	char stato[FD_STS_SIZE];
	char s_dma;
	int drv;
	des_fd *p_des=&fd;
	dma_reg *d_des=&cdma[0];
	
	for (;;) {
		drv=p_des->curr_drv;	// C'e' la possibilita' che arrivino
		p_des->curr_drv=-1;	//  interrupt non richiesti
		if (drv!=-1) {
			if (p_des->curr_cmd!=READ && p_des->curr_cmd!=WRITE) {
			// INT prodotte da comandi di seek o recalibrate
					p_des->drive[drv].errore=fdc_sensei(drv);
			}
			else {
				fdc_status(stato);	// Risposta corretta
				motore_off(p_des->indreg.iDOR,p_des->drive[drv].motore);
				halt_moder(d_des->iREQREG,d_des->iMASK,2);
				if ((stato[0]&0xC0)!=0x00)
					p_des->drive[drv].errore=FD_ERR_NOTX;
				else {
					p_des->drive[drv].errore=D_ERR_NONE;
					p_des->drive[drv].traccia=stato[3];
				}
				dma_status(d_des->iCMD_STS,s_dma);
			}
			sem_signal(p_des->sincr);	// In ogni caso sveglia
		}					//  il processo
		else
			fdc_sensei(-1);		// Interruzione da ignorare
		nwfi(master);			// EOI al master PIC
	}
}
	
const int FDC_IRQ=6;

// Inizializzazione del contollore floppy; max 1 controllore e 2 drive da 1.44MB
// 
void fdc_init() {
	char stato;
	int n_drive;
	char buff[FD_STS_SIZE];

	des_fd *p_des;
	
	short id;
	bool r1,r2;
	
	p_des=&fd;

	printk("Inizializzazione Floppy drive.... \n");
	p_des->drive[0].presente=true;
	n_drive=test_fd();		// Controlla la presenza di uno o piu'
	if (n_drive==1)			//  drive dalla CMOS
		p_des->drive[1].presente=false;
	else
		p_des->drive[1].presente=true;
	outputb(0x00,p_des->indreg.iDOR);	// Disabilita il controllore
	outputb(0x00,p_des->indreg.iD_RATE);	// Data Rate	
	outputb(0x0C,p_des->indreg.iDOR);	// Riabilita il controllore
	p_des->drive[0].motore=false;		//  i motori sono spenti
	p_des->drive[1].motore=false;
	invia_cmd_fd(p_des,SPECIFY);		// Imposta i parametri del floppy
	outputb_fd(0xDF);			// parametri secondari
	outputb_fd(0x06);			//  HeadLoadTime=6ms, DMA=si
	for (int d=0;d<n_drive && p_des->drive[d].presente;d++) {
		motore_on(p_des->indreg.iDOR,p_des->drive[d].motore,d);
		/*	RECALIBRATE	*/
		int i=0;		// Reimplementata per funzionare in
		do {			//  polling (int. sono disabilitate)
			invia_cmd_fd(p_des,RECALIBRATE);
			outputb_fd(d);
			stato=fdc_sensei(d);
			fdc_delay();	// Si rinuncia se non si riesce
		} while ((stato&0x30)==0x30 && (++i)<=FD_RECAL_TIMEOUT);
		/*	----------	*/
		motore_off(p_des->indreg.iDOR,p_des->drive[d].motore);
	}
	p_des->curr_drv=-1;
	p_des->dma = 2;
	
	printk("Assegnato indirizzo DMA %d \n",p_des->dma);
	p_des->dmareg.iMAR=(char*)dma_chan_ind[p_des->dma][0] ;
	p_des->dmareg.iTCR=(char*)dma_chan_ind[p_des->dma][1] ;
	p_des->dmareg.iPAGE=(char*)dma_chan_ind[p_des->dma][2] ;
	p_des->dmareg.iCLEAR_FF=(char*)dma_chan_ind[p_des->dma][3] ;			
//	printk("Indirizzi relativi MAR %x TCR %x ",p_des->dmareg.iMAR,p_des->dmareg.iTCR);
//	printk("PAGE %x CLEAR_FF %x\n",p_des->dmareg.iPAGE,p_des->dmareg.iCLEAR_FF);
	sem_ini(p_des->mutex, 1, r1);
	sem_ini(p_des->sincr, 0, r2);
	if(!r1 || !r2)
		panic("\nImpossibile allocare i semafori per l' IO");
	activate_pe(dmaest_fd,0,PRIO_ESTERN_BASE+IRQ_MAX-FDC_IRQ,LIV_SISTEMA,id,floppy,r1);
	if (!r1 || !r2)
		panic("\nImpossibile creare un processo esterno del floppy disk");
	abilita_fd();		// Abilita le interruzioni dal floppy
}

////////////////////////////////////////////////////////////////////////////////
//               		GESTIONE DELL'HARD DISK               	      //
////////////////////////////////////////////////////////////////////////////////

#define STRICT_ATA

#define HD_MAX_TX 256

#define HD_STS_ERR 0x01		//Maschere per i biti fondamentali dei registri
#define HD_STS_DRQ 0x08
#define HD_STS_DRDY 0x40
#define HD_STS_BSY 0x80
#define HD_DRV_MASTER 0
#define HD_DRV_SLAVE 1

const int A=2;

enum hd_cmd {			// Comandi per i controllori degli hard disk
	NONE=0x00,
	IDENTIFY=0xEC,
	SET_FEATURES=0xEF,
	SET_MULT_MODE=0xC6,
	WRITE_SECT=0x30,
	READ_SECT=0x20,
	HD_SEEK=0x70,
	DIAGN=0x90
};

struct interfata_reg {
	union {				// Command Block Register
		ind_b iCMD;
		ind_b iSTS;
	}; //iCMD_iSTS;
	ind_w iDATA;
	union {
		ind_b iFEATURES;
		ind_b iERROR;
	}; //iFEATURES_iERROR
	ind_b iSTCOUNT;
	ind_b iSECT_N;
	ind_b iCYL_L_N;
	ind_b iCYL_H_N;
	union {
		ind_b iHD_N;
		ind_b iDRV_HD;
	}; //iHD_N_iDRV_HD
	union {				// Control Block Register
		ind_b iALT_STS;		
		ind_b iDEV_CTRL;
	}; //iALT_STS_iDEV_CTRL
};

struct geometria {			// Struttura 3-D di un disco
	unsigned short cil;	
	unsigned short test;
	unsigned short sett;
	unsigned int tot_sett;		// Utile in LBA
};

struct drive {				// Drive su un canale ATA
	bool presente;
	bool dma;
	geometria geom;
};

struct des_ata {		// Descrittore operazione per i controller ATA
	interfata_reg indreg;
	drive disco[2];		// Ogni controller ATA gestisce 2 drive
	hd_cmd comando;
	char errore;
	char cont;
	ind_w punt;	
	int mutex;
	int sincr;
};

extern "C" des_ata hd[A];	// 2 controller ATA

const estern_id hd_id[2]={ata0,ata1};

//Funzioni di utilita' per la gestione del disco
//
inline bool hd_drq(char &s) {		// Test di Data ReQuest
	return s&HD_STS_DRQ;
}

inline bool hd_errore(char &s) {	// Test di un eventuale errore (generico)
	return s&HD_STS_ERR;
}

inline bool hd_drdy(char &s) {		// Test di Drive ReaDY
	return s&HD_STS_DRDY;
}

// Attende una particolare configurazione di bit nel registro di stato del
//  controller; tipicamente deve resettarsi BuSY
//  
char aspetta_STS(des_ata *p_des,char bit,bool &ris) {
	char stato,test;
	short timeout=D_TIMEOUT;		// Previsto un timeout
	do {
		inputb(p_des->indreg.iALT_STS,stato);
		if (bit==HD_STS_DRDY || bit==HD_STS_DRQ)
			test=(~stato);
		else
			test=stato;
		test&=bit;
		timeout--;
		if (timeout<0) {
			ris=false;		// Scaduto il timeout!
			return stato;
		}
	} while (test!=0);
	ris=true;
	return stato;
}

// Invia un comando al controller con la temporizzazione corretta
// 
inline bool invia_cmd(des_ata *p_des,hd_cmd com) {
#ifdef STRICT_ATA
	bool ris=true;
	aspetta_STS(p_des,HD_STS_DRDY,ris);
	if (!ris)
		return false;
#endif
	p_des->comando=com;
	outputb(com,p_des->indreg.iCMD);
	return true;
}

// Legge il buffer di uscita del controller se questo contiene dati
// 
void leggi_sett_buff(des_ata *p_des,short vetti[],char &stato) {
	if (!hd_drq(stato))		
		// DRQ _deve_ essere settato!
		panic("\nErrore Grave: DRQ non e' settato!");
	inputbuffw(p_des->indreg.iDATA,vetti,H_BLK_SIZE);
	//for (int i=0;i<H_BLK_SIZE;i++) //<=== Forma equivalente (debugging)
	//	inputw(p_des->indreg.iDATA,vetti[i]);
}

// Scrive il buffer di ingresso del controller se questo aspetta dati
// 
void scrivi_sett_buff(des_ata *p_des,short vetti[],char &stato) {
	if (!hd_drq(stato))		
		// DRQ _deve_ essere settato!
		panic("\nErrore Grave: DRQ non e' settato!");
	outputbuffw(vetti,p_des->indreg.iDATA,H_BLK_SIZE);
	//for (int i=0;i<H_BLK_SIZE;i++) //<=== Forma equivalente (debugging)
	//	outputw(vetti[i],p_des->indreg.iDATA);
}

//Funzioni per l'ingresso/uscita dati e processo estern0
//
extern "C" void go_inouthd(ind_b i_dev_ctl);

extern "C" void halt_inouthd(ind_b i_dev_ctl);

extern "C" bool test_canale(ind_b st,ind_b stc,ind_b stn);

extern "C" int leggi_signature(ind_b stc,ind_b stn,ind_b cyl,ind_b cyh);

extern "C" void abilita_hd(ind_b h);

extern "C" bool drive_reset(ind_b dvctrl,ind_b st);

extern "C" void set_drive(short ms,ind_b i_drv_hd);

extern "C" void get_drive(ind_b i_drv_hd,short &ms);

extern "C" void setup_addr_hd(des_ata *p,unsigned int primo);	

// Trasferimento dati in PIO Mode
//

// Avvio delle operazioni di lettura da hard dik
// 
void starthd_in(des_ata *p_des,short drv,short vetti[],unsigned int primo,unsigned char &quanti)	{
	p_des->punt=vetti;
#ifdef STRICT_ATA
	bool ris;
	aspetta_STS(p_des,HD_STS_BSY,ris);	// aspetta "drive not BuSY"
#endif
	set_drive(drv,p_des->indreg.iDRV_HD);	// Selezione del drive richiesto
#ifdef STRICT_ATA
	char stato=aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta di nuovo!!
	if (!hd_drdy(stato) || hd_errore(stato))
			panic("Errore nell'inizializzazione della lettura!");
#endif
	setup_addr_hd(p_des,primo); 
	outputb(quanti,p_des->indreg.iSTCOUNT);
	// Abilita le interruzioni per il controller-drive selezionato
	go_inouthd(p_des->indreg.iDEV_CTRL);	
	invia_cmd(p_des,READ_SECT);
}

// Avvio delle operazioni di scrittura su hard disk
//
void starthd_out(des_ata *p_des,short drv,short vetti[],unsigned int primo,unsigned char &quanti)	
{	char stato;				
	bool ris;
	p_des->punt=vetti;
#ifdef STRICT_ATA
	aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta "drive not BuSY"
#endif
	set_drive(drv,p_des->indreg.iDRV_HD);	// Selezione del drive richiesto
#ifdef STRICT_ATA
	stato=aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta di nuovo!
	if (!hd_drdy(stato) || hd_errore(stato))
			panic("Errore nell'inizializzazione della scrittura!");
#endif
	setup_addr_hd(p_des,primo); 
	outputb(quanti,p_des->indreg.iSTCOUNT);
	//Abilita le interruzioni per il controller-drive selezionato
	go_inouthd(p_des->indreg.iDEV_CTRL);		
	invia_cmd(p_des,WRITE_SECT);	// Asimmetria con la lettura, bisogna
#ifdef STRICT_ATA			//  scrivere il primo blocco
	aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta "drive not BuSY"
#endif
	// Aspetta DRQ, previsto dal protocollo di comunicazione
	stato=aspetta_STS(p_des,HD_STS_DRQ,ris);
	lock();
	inputb(p_des->indreg.iSTS,stato);
	scrivi_sett_buff(p_des,p_des->punt,stato);
	p_des->punt+=H_BLK_SIZE;
	unlock();
}

// Primitiva di lettura da hard disk: legge quanti blocchi (max 256, char!)
//  a partire da primo depositandoli in vetti; il controller usato e ind_ata,
//  il drive e' drv. Riporta un fallimento in errore
//  
extern "C" void c_readhd_n(short ind_ata,short drv,short vetti[],unsigned int primo,unsigned char &quanti,char &errore)
{
	des_ata *p_des;

	p_des = &hd[ind_ata];

	// Controllo sulla protezione
	if (ind_ata<0 || ind_ata>=A || drv<0 || drv>=2)
		return;
	if (!verifica_area(&quanti,sizeof(int),true) || !verifica_area(vetti,quanti,true) || !verifica_area(&errore,sizeof(char),true))
		return;

	// Controllo sulla selezione di un drive presente
	if (p_des->disco[drv].presente==false) {
		errore=D_ERR_PRESENCE;
		return;
	}

	// Controllo sull'indirizzamento
	if (primo>p_des->disco[drv].geom.tot_sett)
		errore=D_ERR_BOUNDS;
	else {		
		sem_wait(p_des->mutex);
		p_des->errore=D_ERR_NONE;		// Reset degli errori precedenti 
		p_des->disco[drv].dma=false;	// Trasferimento in PIO Mode
		// Abilitazione del controller all'operazione
		starthd_in(p_des,drv,vetti,primo,quanti);
		sem_wait(p_des->sincr);
		quanti=p_des->cont;
		errore=p_des->errore;
		sem_signal(p_des->mutex);
	}
}

// Primitiva di scrittura su hard disk: scrive quanti blocchi (max 256, char!)
//  a partire da primo prelevandoli da vetti; il controller usato e' ind_ata,
//  il drive e' drv. Riporta un fallimento in errore
//  
extern "C" void c_writehd_n(short ind_ata,short drv,short vetti[],unsigned int primo,unsigned char &quanti,char &errore)
{
	des_ata *p_des;

	p_des = &hd[ind_ata];

	// Controllo sulla protezione
	if (ind_ata<0 || ind_ata>=A || drv<0 || drv>=2)
		return;
	if (!verifica_area(&quanti,sizeof(int),true) || !verifica_area(vetti,quanti,true) || !verifica_area(&errore,sizeof(char),true))
		return;

	// Controllo sulla selezione di un drive presente
	if (p_des->disco[drv].presente==false) {
		errore=D_ERR_PRESENCE;
		return;
	}
	// Controllo sull'indirizzamento
	if (primo>p_des->disco[drv].geom.tot_sett)
		errore=D_ERR_BOUNDS;
	else {
		sem_wait(p_des->mutex);
		p_des->errore=D_ERR_NONE;		// Reset degli errori precedenti
		p_des->disco[drv].dma=false;	//Trasferimento in PIO Mode
		starthd_out(p_des,drv,vetti,primo,quanti);
		sem_wait(p_des->sincr);
		quanti=p_des->cont;
		errore=p_des->errore;
		sem_signal(p_des->mutex);
	}
}

// Processo esterno per l'hard disk, valido sia per l'ingresso che per 
//  l'uscita
//
void estern_hd(int h)
{
	des_ata *p_des;
	char stato;
	short drv;
	hd_cmd curr_cmd;
	bool fine;

	p_des = &hd[h];		// In base al parametro si sceglie un controller

	for(;;) {
		curr_cmd=p_des->comando;	// Gestione delle interruzioni
		p_des->comando=NONE;		//  non richieste
		if (curr_cmd!=NONE) {
			get_drive(p_des->indreg.iDRV_HD,drv);
#ifdef STRICT_ATA
			bool ris;
			stato=aspetta_STS(p_des,HD_STS_BSY,ris);
#endif	/* STRICT_ATA */
			fine=false;
			inputb(p_des->indreg.iSTS,stato);
			if (!hd_errore(stato) && (curr_cmd==READ_SECT || curr_cmd==WRITE_SECT)) {
				if (curr_cmd==WRITE_SECT) {
					inputb(p_des->indreg.iSTCOUNT,p_des->cont);
					if (p_des->cont==0)
						fine=true;
					else {
						scrivi_sett_buff(p_des,p_des->punt,stato);
						p_des->comando=WRITE_SECT;
					}
				}
				else {		// Comando di READ_SECT
					leggi_sett_buff(p_des,p_des->punt,stato);
					inputb(p_des->indreg.iSTCOUNT,p_des->cont);
					if (p_des->cont==0)
						fine=true;
					else
						p_des->comando=READ_SECT;
				}
				p_des->punt+=H_BLK_SIZE;
			}
			else {
				inputb(p_des->indreg.iERROR,p_des->errore);
				fine=true;
			}
			if (fine==true) {
				//Disabilita le interruzioni dal controller
				halt_inouthd(p_des->indreg.iDEV_CTRL);	
				if (curr_cmd!=IDENTIFY && curr_cmd!=DIAGN)
					sem_signal(p_des->sincr);
			}	
		}
		nwfi(slave); 	// Controller ATA = Slave PIC (entrambi, 14 e 15)
	}
}

const int ATA1_IRQ=15;

// Inizializzazione e autoriconoscimento degli hard disk collegati ai canali ATA
// 
void hd_init() {
	char stato=0;
	bool test;
	int init_time=D_TIMEOUT;
	short st_sett[H_BLK_SIZE];
	short id;
	bool r,r1,r2;
	int hd_base_prio=PRIO_ESTERN_BASE+IRQ_MAX-ATA1_IRQ;
	short canali=0;
	
	des_ata *p_des;
	for (int i=0;i<A;i++) {				// Per ogni controller
		p_des=&hd[i];
		// Disabilita le interruzioni
		halt_inouthd(p_des->indreg.iDEV_CTRL);	
	}
	for (int i=0;i<A;i++) {			// Per ogni controller
		p_des=&hd[i];
		short drv=HD_DRV_MASTER;
		set_drive(drv,p_des->indreg.iDRV_HD);	//LBA e drive drv
		// Procedura diversificata una volta effettuato il test()
		if (test_canale(p_des->indreg.iSTS,p_des->indreg.iSTCOUNT,p_des->indreg.iSECT_N)) {
#ifndef BOCHS		// In sistemi reali e' necessario inviare questo comando
			set_drive(HD_DRV_MASTER,p_des->indreg.iDRV_HD);
			invia_cmd(p_des,DIAGN);		// Diagnostica dei drive
			do
				inputb(p_des->indreg.iSTS,stato);
			while (stato&HD_STS_BSY);	// Polling
			inputb(p_des->indreg.iERROR,stato);
			do
				inputb(p_des->indreg.iSTS,stato);
			while (!stato&HD_STS_DRDY);	// Polling
#endif		// Diagnostica terminata

			for (int d=0;d<2;d++) {		// Per ogni drive
				set_drive(drv,p_des->indreg.iDRV_HD);
				p_des->disco[d].dma=false;
#ifdef BOCHS		// BOCHS ha bisogno del reset dei controller
				if (!drive_reset(p_des->indreg.iDEV_CTRL,p_des->indreg.iSTS)) {
					p_des->disco[d].presente=false;
					drv=HD_DRV_SLAVE;
					continue;
				}
				else
					p_des->disco[d].presente=true;
#endif		// A questo punto in entrambi i casi e' disponibile la signature
				if (leggi_signature(p_des->indreg.iSTCOUNT,p_des->indreg.iSECT_N,p_des->indreg.iCYL_L_N,p_des->indreg.iCYL_H_N)!=0) {
					p_des->disco[d].presente=false;
					drv=HD_DRV_SLAVE;
					continue;
				}
				else
					p_des->disco[d].presente=true;
		
		//Riconoscimento della geometria e dei parametri del disco
				if (!invia_cmd(p_des,IDENTIFY)) {	
					p_des->disco[d].presente=false;
					drv=HD_DRV_SLAVE;
					continue;
				}
#ifdef STRICT_ATA
				bool ris;
				aspetta_STS(p_des,HD_STS_BSY,ris);
#endif
				inputb(p_des->indreg.iSTS,stato);
				leggi_sett_buff(p_des,st_sett,stato);
				// Inizializzazione della geometria
				p_des->disco[d].geom.cil=st_sett[1];		
				p_des->disco[d].geom.test=st_sett[3];
				p_des->disco[d].geom.sett=st_sett[6];
				p_des->disco[d].geom.tot_sett=(st_sett[58]<<16)+st_sett[57];
				drv=HD_DRV_SLAVE;
			}
			sem_ini(p_des->mutex, 1, r1);
			sem_ini(p_des->sincr, 0, r2);
			if(!r1 || !r2)
				panic("\nImpossibile allocare i semafori per l' IO");
			activate_pe(estern_hd,i,hd_base_prio-i,LIV_SISTEMA,id,hd_id[i],r);
			if(!r)
				panic("\nImpossibile creare un processo esterno dell'hard disk");
		}
		else
			p_des->disco[0].presente=p_des->disco[1].presente=false;
		p_des->comando=NONE;
		abilita_hd(p_des->indreg.iSTS);
	}
}

// Funzione di utilita' perche' l'utente possa conoscere se interessato,
//  le dimensioni e la struttura dei suoi hard disk
//  
extern "C" void c_geometria(int interf,short drv,unsigned short &c,unsigned short &t,unsigned short &s,int &ts,char &err) {
	if (hd[interf].disco[drv].presente==true) {
		c=hd[interf].disco[drv].geom.cil;
		t=hd[interf].disco[drv].geom.test;
		s=hd[interf].disco[drv].geom.sett;
		ts=hd[interf].disco[drv].geom.tot_sett;
		err=D_ERR_NONE;
	}
	else
		err=D_ERR_PRESENCE;
}

//////////////////////////////////////////////////////////////////////////////
//       SUPPORTO AI RIMPIAZZAMENTI DELLE PAGINE IN MEMORIA DI MASSA        //
//////////////////////////////////////////////////////////////////////////////

#define PAGE_SIZE	4096

// Se out=true: sposta QUANTE pagine di memoria a partire da quella puntata
//  da PAG (come short*) sul disco DRV del controller IND_ATA a partire dal
//  PRIMO blocco
// Se out=false: sposta QUANTE pagine di memoria alla locazione puntata
//  da PAG (come short*) prelevandole dal disco DRV del controller IND_ATA
//  in posizione PRIMO 
//  
extern "C" void c_swap_pag(bool out,short ind_ata,short drv,unsigned int primo,void* pag,int quante) {
	des_ata *p_des;
	bool ris;
	char stato;

	p_des = &hd[ind_ata];

	p_des->punt=(short*)pag;
#ifdef STRICT_ATA
	aspetta_STS(p_des,HD_STS_BSY,ris);	// aspetta "drive not BuSY"
#endif
	set_drive(drv,p_des->indreg.iDRV_HD);	// Selezione del drive richiesto
#ifdef STRICT_ATA
	stato=aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta di nuovo!!
	if (!hd_drdy(stato) || hd_errore(stato))
			panic("Errore durante un rimpiazzamento");
#endif
	// Disabilita le interruzioni per il controller/drive selezionato
	halt_inouthd(p_des->indreg.iDEV_CTRL);

	quante=(quante*PAGE_SIZE)/BLK_SIZE;
	do {
		setup_addr_hd(p_des,primo);
		outputb((quante<HD_MAX_TX?quante:0),p_des->indreg.iSTCOUNT);
		quante-=HD_MAX_TX;
		if (!out)
			invia_cmd(p_des,READ_SECT);
		else
			invia_cmd(p_des,WRITE_SECT);
		do {		
#ifdef STRICT_ATA
			aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta "drive not BuSY"
#endif
			// Aspetta DRQ, previsto dal protocollo di comunicazione
			stato=aspetta_STS(p_des,HD_STS_DRQ,ris);
			inputb(p_des->indreg.iSTS,stato);
			if (!out)
				leggi_sett_buff(p_des,p_des->punt,stato);
			else
				scrivi_sett_buff(p_des,p_des->punt,stato);	
			inputb(p_des->indreg.iSTS,stato);
			if (hd_errore(stato))
				panic("Errore durante un rimpiazzamento");
			p_des->punt+=H_BLK_SIZE;
			inputb(p_des->indreg.iSTCOUNT,p_des->cont);
		} while ((p_des->cont)!=0);
		primo+=HD_MAX_TX;
	} while (quante>0);
}
////////////////////////////////////////////////////////////////////////////////
//                 FUNZIONI SCHEDA AUDIO COMPATIBILE SB16                     //
////////////////////////////////////////////////////////////////////////////////
//porte del mixer
#define MASTERLEFT  0x30
#define MASTERRIGHT 0x31
#define VOCLEFT 0x32
#define VOCRIGHT 0x33
#define MIDILEFT 0x34
#define MIDIRIGHT 0x35
#define CDLEFT 0x36
#define CDRIGHT 0x37
#define LINELEFT 0x38
#define LINERIGHT 0x39
#define MICVOLUME 0x3a
#define PCSPEAKER 0x3b
#define OUTPUTSWITCHES 0x3c
#define INPUTSWITCHESLEFT 0x3d
#define INPUTSWITCHESRIGHT 0x3e
#define INPUTGAINLEFT 0x3f
#define INPUTGAINRIGHT 0x40
#define OUTPUTGAINLEFT 0x41
#define OUTPUTGAINRIGHT 0x42
#define AGC 0x43
#define TREBLELEFT 0x44
#define TREBLERIGHT 0x45
#define BASSLEFT 0x46
#define BASSRIGHT 0x47

#define SB_ERR_BOUNDS 0xFF

#define SB_ERR_DMA 0xFA

#define BLK_SIZE 512		

#define SB_N_BLK 2880
#define SB_MAX_BLK 128



//********************************
//*** Descrittore 
//********************************

//********************************
//*** Registri DSP
//********************************
struct sbcontr_reg
{	
	union {
		ind_b FM_STS_1;
		ind_b FM_REG_1;
	}; //base+0
	ind_b FM_DATA_1; //base+1
	union {
		ind_b FM_AD_STS;
		ind_b FM_AD_REG;
	}; //base+2
	ind_b FM_AD_DATA; //base+3
	ind_b MIX_REG; //base+4
	ind_b MIX_DATA; //base+5
	ind_b DSP_RESET; //base+6
	ind_b RES_1; //base+7
	union { 
		ind_b FM_STS_2;
		ind_b FM_REG_2;
	}; //base+8
	ind_b FM_DATA_2; //base+9
	ind_b DSP_READ; //base+a
	ind_b RES_2; //base+b
	union { 
		ind_b DSP_WRITE;
		ind_b DSP_WRITE_STS;
	}; //base+c
	ind_b RES_3; //base+d
	ind_b DSP_READ_STS; //base+e
	ind_b RES_4; //base+f
	ind_b CD_DATA; //base+10
	ind_b CD_STS; //base+11
	ind_b CD_RESET; //base12
	ind_b CD_ENABLE; //base+13
	union {
		ind_b MIDI_STS;
		ind_b MIDI_REG;
	}; 
	ind_b MIDI_DATA;
};
//********************************
//*** struttura descrittore
//********************************
struct des_sb {
	sbcontr_reg indreg;
	dmachan_reg dmareg;
	int mutex;
	int sincr;
	int modebits;
	int modestereomono;
	int soundrate;
	int transferlength;
	int dma;
	int dma16;	
	int irq;
	int transfermode;
};
#define BUFFSIZE        8192
#define HALFBUFFSIZE    4096

extern "C" des_sb sb_16;

extern "C" void abilita_sb(char irq);

/////////////////////////////////////////////////////////////////////
// ******************   FUNZIONI DI ACCESSO GLOBALI *****************
/////////////////////////////////////////////////////////////////////
  
//funzione di lettura dati da scheda audio 
char read_dsp()
  {
  des_sb* p_des=&sb_16;
  char a;
  inputb(p_des->indreg.DSP_READ_STS,a);
   while (!((unsigned char)a & 0x80))
     {   inputb(p_des->indreg.DSP_READ_STS,a);}
  inputb(p_des->indreg.DSP_READ,a);

    return(a);
  }
  
//funzione scrittura dati-comandi scheda audio
void write_dsp(char value)
  {
  des_sb *p_des=&sb_16;
  char a;
  inputb(p_des->indreg.DSP_WRITE,a);
   while ((unsigned char)a  & 0x80)
      {inputb(p_des->indreg.DSP_WRITE,a); }
    outputb( (unsigned char)value,p_des->indreg.DSP_WRITE);
  }

//accensione speaker
void speakerOn(void){
	write_dsp(0xD1);
}
//spegnimento speaker
void speakerOff(void){
	write_dsp(0xD3);
}
//Reset DSP
bool reset_dsp(int baseio) //occore specificare baseio poich  chiammata prima di aver inizializzato il descrittore
{
int a;
unsigned char b;
char aa;
int c;
    outputb( 1,(char*)(baseio+0x006));
    while (a < 300) {a++;}
    outputb( 0,(char*)(baseio+0x006));	
    c = 0xFFF;
    inputb((char*)(baseio+0x00E),aa);
    while (!((unsigned char)aa & 0x80) && (c > 0))
     {
        inputb((char*)(baseio+0x00E),aa);
     	 c--;
     }
     if (c <= 0) return 0;
    c = 0xFFF;
    inputb((char*)(baseio+0x00A),aa);
    while (((unsigned char)aa != 0xaa) && (c > 0) ) 
    {	inputb((char*)(baseio+0x00A),aa);
    	c--;
    }
    if (c <= 0) return 0;
    return 1;
}
//Fermo trasferimenti DMA 
void dsp_stopall(void)
  {
    des_sb *p_des=&sb_16;
   // pause 8/16-bit DMA mode digitized sound I/O
    reset_dsp((int)p_des->indreg.FM_STS_1);
    write_dsp(0xD0);
    write_dsp(0xD5);
  }
//Ritorna lo stato degli speaker 0xFF = ON
unsigned char getSpeakerStatus(void){
	write_dsp(0xD8);
	return read_dsp();
}
//primitive offerte per l'utente
extern "C" void c_getSpeakerStatus(unsigned char &status){
	status=getSpeakerStatus();
}
extern "C" void c_speakerOn(void){
	speakerOn();
}  
extern "C" void c_speakerOff(void){
	speakerOff();
}
/////////////////////////////////////////////////////////////////////
// ******************   FUNZIONI GESTIONE MIXER  ********************
/////////////////////////////////////////////////////////////////////

//funzione lettura dati da mixer
unsigned char getMixerSetting(char port)
{
 char a;
  des_sb *p_des=&sb_16;
 outputb(port,p_des->indreg.MIX_REG);
 inputb(p_des->indreg.MIX_DATA,a);

 if(port >= MASTERLEFT && port <=MICVOLUME)
  {return (unsigned char)a>>3;
  }
 else
  if(port ==PCSPEAKER)
  {return (unsigned char)a>>6;
  }
 else 
  if(port >=OUTPUTSWITCHES && port <=INPUTSWITCHESRIGHT)
  {return (unsigned char)a;
  }
 else 
  if(port >=INPUTGAINLEFT && port <=INPUTGAINRIGHT)
   {return (unsigned char)a>>6;
   }
 else
  if(port ==AGC)
  {return (unsigned char)a;
  }
 else
  if(port >=TREBLELEFT && port <=BASSRIGHT)
  {return (unsigned char)a>>4;
  }
 else return (unsigned char)a;
}

//funzione scrittura dati mixer
void setMixerSetting(char port, char value)
{
  des_sb *p_des=&sb_16;
 if(port >=MASTERLEFT && port <=MICVOLUME)
  {outputb(port,p_des->indreg.MIX_REG);
    outputb(value<<3,p_des->indreg.MIX_DATA);
  }
 else
  if(port ==PCSPEAKER)
  { outputb(port,p_des->indreg.MIX_REG);
    outputb(value<<6,p_des->indreg.MIX_DATA);
  }
 else 
  if(port >=OUTPUTSWITCHES && port <=INPUTSWITCHESRIGHT)
  { outputb(port,p_des->indreg.MIX_REG);
    outputb(value,p_des->indreg.MIX_DATA);
  }
 else 
  if(port >=INPUTGAINLEFT && port <=INPUTGAINRIGHT)
   { outputb(port,p_des->indreg.MIX_REG);
    outputb(value<<6,p_des->indreg.MIX_DATA);
   }
 else
  if(port ==AGC)
  { outputb(port,p_des->indreg.MIX_REG);
    outputb(value,p_des->indreg.MIX_DATA);
  }
 else
  if(port >=TREBLELEFT && port <=BASSRIGHT)
  { outputb(port,p_des->indreg.MIX_REG);
    outputb(value<<4,p_des->indreg.MIX_DATA);
  }
 }
 //primitive offerte per l'utente
extern "C" void c_setAllMixerSetting(char vetti[])
{
 int i;
 for (i=MASTERLEFT;i<=BASSRIGHT;i++)
 	setMixerSetting((char)i,vetti[i-MASTERLEFT]);
} 

extern "C" void c_getAllMixerSetting(char vetti[])
{
 int i;
 for (i=MASTERLEFT;i<=BASSRIGHT;i++)
 	vetti[i-MASTERLEFT]=getMixerSetting((char)i);
}

extern "C" void c_setMixerSetting(char port, char value){
	if ((port>MASTERLEFT) && (port<=BASSRIGHT))
		setMixerSetting(port,value);
}

extern "C" void c_getMixerSetting(char port,unsigned char value){
	if ((port>MASTERLEFT) && (port<=BASSRIGHT))
		value=getMixerSetting(port);
}
/////////////////////////////////////////////////////////////////////
// ******************   FUNZIONI MIDI MPU401  ***********************
/////////////////////////////////////////////////////////////////////

//lettura dati midi
char read_midi()
  {
  des_sb* p_des=&sb_16;
  char a;
  inputb(p_des->indreg.MIDI_STS,a);
   while (((unsigned char)a & 0x80))
     {   inputb(p_des->indreg.MIDI_STS,a);}
  inputb(p_des->indreg.MIDI_DATA,a);

    return(a);
  }
  
//scrittura dati midi
void write_midi(char value)
  {
  des_sb *p_des=&sb_16;
  char a;
  inputb(p_des->indreg.MIDI_STS,a);
   while (((unsigned char)a  & 0x40))
      {inputb(p_des->indreg.MIDI_STS,a); }
    outputb( (unsigned char)value,p_des->indreg.MIDI_DATA);
  }

#define MIDI_RESET 0xff
#define MIDI_UART 0x3f
//scrittura comandi midi
bool cmd_midi(unsigned char value)
  {
  des_sb *p_des=&sb_16;
  char a;
  int i;
  inputb(p_des->indreg.MIDI_STS,a);
   while (((unsigned char)a  & 0x40))
      {inputb(p_des->indreg.MIDI_STS,a);
       }
   outputb(value,p_des->indreg.MIDI_REG);
   inputb(p_des->indreg.MIDI_STS,a);
   i=0;
   while (((unsigned char)a  & 0x80) && (i<65536))
      {inputb(p_des->indreg.MIDI_STS,a);
      i++;
      }
   if (!(i<65536)) {
   	return true;}
   else {
   	inputb(p_des->indreg.MIDI_DATA,a);
	if ((unsigned char)a == 0xFE)
		return true;
	else
		return false;
   }
    
  }
//primitive offerte per l'utente
extern "C" void c_readMidi(unsigned char value){
	value=(unsigned char)read_midi();
}

extern "C" void c_writeMidi(unsigned char &value){
	write_midi(value);
}

extern "C" void c_resetMidi(bool &res){
	res=cmd_midi(MIDI_RESET);
}

extern "C" void c_startMidi(bool &res){
	res=cmd_midi(MIDI_UART);
}


/////////////////////////////////////////////////////////////////////
// ******************   PROCESSO ESTERNO         ********************
/////////////////////////////////////////////////////////////////////

// Processo esterno 
// 
void dmaest_sb(int h) {
	char s_dma,a;
	des_sb *p_des=&sb_16;
	dma_reg *d_des=&cdma[0];
	for (;;) {
		halt_moder(d_des->iREQREG,d_des->iMASK,1); //fermo DMA
		if (p_des->transfermode == 1) //ciclo singolo
			{
			inputb(p_des->indreg.DSP_READ_STS,a); // ACK alla SB
   			}	
		dma_status(d_des->iCMD_STS,s_dma); //ACK al DMA
		sem_signal(p_des->sincr);	// Rilascio il semaforo
		speakerOff();			// Spengo gli speaker
		nwfi(master);			// EOI al master PIC
	}
}

/////////////////////////////////////////////////////////////////////
// ******************   FUNZIONI I/O IN DMA  ************************
/////////////////////////////////////////////////////////////////////

#define LoByte(x)(short)(x & 0x00FF)
#define HiByte(x)(short)((x & 0x00FF) >> 8)
//primitive offerte per l'utente
void imposta_suono(des_sb *p_des,int vetti[],int quanti)
{
	p_des->modebits=vetti[0];
	p_des->modestereomono=vetti[1];
	p_des->soundrate=vetti[2];
	p_des->transfermode=vetti[4];
	p_des->transferlength=BUFFSIZE;
}

//Funzione invio lunghezza trasferimento
void sb_send_length(int transferlength)
{
	write_dsp(LoByte(transferlength-1));
	write_dsp(HiByte(transferlength-1));
}
//funzione invio frequenza
void sb_send_frequency(int soundrate)
{
  	write_dsp(0x41);
	write_dsp(LoByte(soundrate-1));
	write_dsp(HiByte(soundrate-1));
}
//Funzione Setup DSP per DMA transfer out
void sb_setupDSP(int modestereomono,int modebits,int transfermode,int soundrate,int transferlength)
{
	write_dsp(0xD1); //speakerON
	if (transfermode == 1) 
		{if (modebits == 8)
			{if (modestereomono==0)	
				{//Ciclo singolo Mono 8 bit
				sb_send_frequency(soundrate);
				write_dsp(0xC0);
				write_dsp(0x00);
				sb_send_length(transferlength);
				}
				else
				{//Ciclo singolo Stereo 8 bit
				sb_send_frequency(soundrate);
				write_dsp(0xC0);
				write_dsp(0x20);
				sb_send_length(transferlength);
				}
			}	
		else 
			{if (modestereomono==0)	
				{//Ciclo singolo Mono 16 bit
				sb_send_frequency(soundrate);
				write_dsp(0xc0);
				write_dsp(0x10);
				sb_send_length(transferlength);
				}
				else
				{//Ciclo singolo Stereo 16 bit
				sb_send_frequency(soundrate);
				write_dsp(0xc0);
				write_dsp(0x30);
				sb_send_length(transferlength);
				}	
			}
		}
	else //ciclo AI
		{if (modebits == 8)
			{if (modestereomono==0)	
				{//Ciclo AI Mono 8 bit
				sb_send_frequency(soundrate);
				write_dsp(0xC6);
				write_dsp(0x00);
				sb_send_length(transferlength);
				}
				else
				{//Ciclo AI Stereo 8 bit
				sb_send_frequency(soundrate);
				write_dsp(0xC6);
				write_dsp(0x20);
				sb_send_length(transferlength);
				}
			}	
		else 
			{if (modestereomono==0)	
				{//Ciclo AI Mono 16 bit
				sb_send_frequency(soundrate);
				write_dsp(0xc6);
				write_dsp(0x10);
				sb_send_length(transferlength);
				}
				else
				{//Ciclo AI Stereo 16 bit
				sb_send_frequency(soundrate);
				write_dsp(0xc6);
				write_dsp(0x30);
				sb_send_length(transferlength);
				}	
			}
		}

}
//Funzione setup DSP per SB transfer IN
void sb_setupDSPin(int modestereomono,int modebits,int transfermode,int soundrate,int transferlength)
{
	write_dsp(0xD1); //speaker ON
	if (transfermode == 1) 
		{if (modebits == 8)
			{if (modestereomono==0)	
				{//Ciclo singolo Mono 8 bit
				sb_send_frequency(soundrate);
				write_dsp(0xC8);
				write_dsp(0x00);
				sb_send_length(transferlength);
				}
				else
				{//Ciclo singolo Stereo 8 bit
				sb_send_frequency(soundrate);
				write_dsp(0xC8);
				write_dsp(0x20);
				sb_send_length(transferlength);
				}
			}	
		else 
			{if (modestereomono==0)	
				{//Ciclo singolo Mono 16 bit
				sb_send_frequency(soundrate);
				write_dsp(0xc8);
				write_dsp(0x10);
				sb_send_length(transferlength);
				}
				else
				{//Ciclo singolo Stereo 16 bit
				sb_send_frequency(soundrate);
				write_dsp(0xc8);
				write_dsp(0x30);
				sb_send_length(transferlength);
				}	
			}
		}
	else //ciclo AI
		{if (modebits == 8)
			{if (modestereomono==0)	
				{//Ciclo AI Mono 8 bit
				sb_send_frequency(soundrate);
				write_dsp(0xCe);
				write_dsp(0x00);
				sb_send_length(transferlength);
				}
				else
				{//Ciclo AI Stereo 8 bit
				sb_send_frequency(soundrate);
				write_dsp(0xCe);
				write_dsp(0x20);
				sb_send_length(transferlength);
				}
			}	
		else 
			{if (modestereomono==0)	
				{//Ciclo AI Mono 16 bit
				sb_send_frequency(soundrate);
				write_dsp(0xce);
				write_dsp(0x10);
				sb_send_length(transferlength);
				}
				else
				{//Ciclo AI Stereo 16 bit
				sb_send_frequency(soundrate);
				write_dsp(0xce);
				write_dsp(0x30);
				sb_send_length(transferlength);
				}	
			}
		}
}

//scrittura in DMA 
int sb_dmastart_out(des_sb *p_des,dma_reg *d_des,ind_w vetti, unsigned char &quanti)
{
	ind_fisico iff;
	

	trasforma((ind_l)vetti,iff);		// Ottenimento dell'indirizzo
						//  fisico (DMA!)
	if (iff==0) {
		return SB_ERR_DMA; 
	}
	
	// Inizializzazione del controllore DMA (indirizzi e contatore)
	setup_addr_dma(p_des->dmareg.iPAGE,p_des->dmareg.iMAR,iff);
	setup_cont_dma(p_des->dmareg.iTCR,BLK_SIZE*quanti);
	//Inizializzazione DSP	
	sb_setupDSP(p_des->modestereomono,p_des->modebits,p_des->transfermode,p_des->soundrate,p_des->transferlength);
	go_moder_out(WRITE_BLOCK,d_des->iMODER,d_des->iMASK,p_des->dma);//Programmazione	
	return -1; //Tutto OK
}
	
//lettura in DMA 
int sb_dmastart_in(des_sb *p_des,dma_reg *d_des,ind_w vetti,unsigned char &quanti)
{
	
	ind_fisico iff;
	
	trasforma((ind_l)vetti,iff);		// Ottenimento dell'indirizzo
						//  fisico (DMA!)
	if (iff==0) {
		return SB_ERR_DMA;
	}
	
	// Inizializzazione del controllore DMA (indirizzi e contatore)
	setup_addr_dma(p_des->dmareg.iPAGE,p_des->dmareg.iMAR,iff);
	setup_cont_dma(p_des->dmareg.iTCR,BLK_SIZE*quanti);
	//Inizializzazione DSP
	sb_setupDSPin(p_des->modestereomono,p_des->modebits,p_des->transfermode,p_des->soundrate,p_des->transferlength);
	go_moder_in(READ_BLOCK,d_des->iMODER,d_des->iMASK,p_des->dma);	// Programmazione
	return -1;
}
//calcolo byte rimanenti
inline void get_quantisb(des_sb *p_des,unsigned char &q) {
	char contL,contH;
	outputb(0x00,p_des->dmareg.iCLEAR_FF);
	inputb(p_des->dmareg.iTCR,contL);
	inputb(p_des->dmareg.iTCR,contH);
	q=(contH<<8)+(contL+1);
}
//primitive offerte per l'utente	
extern "C" void c_writesb_n(ind_w vetti,unsigned char &quanti,char &errore,int vett[])
{
	des_sb *p_des;
	dma_reg *d_des;
	char error;
	p_des=&sb_16;
	d_des=&cdma[0];
	// Controllo di protezione
	if (!verifica_area(&quanti,sizeof(int),true) ||
	    !verifica_area(vetti,quanti,true) ||
	    !verifica_area(&errore,sizeof(char),true))
		return;
	
	//Controllo sull'indirizzamento
	if (quanti>SB_N_BLK || quanti>SB_MAX_BLK) {
		errore=SB_ERR_BOUNDS;
		return;
	}
	
	sem_wait(p_des->mutex);
	imposta_suono(p_des,vett,quanti); //imposta parametri trasferimento nel descrittore	
	error=sb_dmastart_out(p_des,d_des,vetti,quanti);
	if (error!=-1) {
		sem_wait(p_des->sincr);
	}
	errore=error;
	get_quantisb(p_des,quanti);
	sem_signal(p_des->mutex);
}


extern "C" void c_readsb_n(ind_w vetti,unsigned char &quanti,char &errore,int vett[])
{
	des_sb *p_des;
	dma_reg *d_des;
	char error;
	p_des=&sb_16;
	d_des=&cdma[0];
	// Controllo di protezione
	if (!verifica_area(&quanti,sizeof(int),true) ||
	    !verifica_area(vetti,quanti,true) ||
	    !verifica_area(&errore,sizeof(char),true))
		return;
	
	//Controllo sull'indirizzamento
	if (quanti>SB_N_BLK || quanti>SB_MAX_BLK) {
		errore=SB_ERR_BOUNDS;
		return;
	}
	
	sem_wait(p_des->mutex);
	imposta_suono(p_des,vett,quanti); //imposta parametri trasferimento nel descrittore	
	error=sb_dmastart_in(p_des,d_des,vetti,quanti);

	if (error!=-1) {
		sem_wait(p_des->sincr);
	}
	errore=error;
	get_quantisb(p_des,quanti);
	sem_signal(p_des->mutex);

}
	

/////////////////////////////////////////////////////////////////////
// ******************   FUNZIONI PER DIRECT PLAY ********************
/////////////////////////////////////////////////////////////////////
// funzione di prova per il direct play
void sb16DirectPlay(unsigned char byte)
{write_dsp(0x10);
 write_dsp(byte);
}

unsigned char sb16DirectRecord(void)
{
	write_dsp(0x20);
	unsigned char b;
	b=(unsigned char)read_dsp();
	return b;
}
//primitive offerte per l'utente
extern "C" void c_DirectPlay(unsigned char byte)
{ 
	sb16DirectPlay(byte);
}

extern "C" void c_DirectRecord(unsigned char &byte)
{ 
	byte = sb16DirectRecord();
}

 /////////////////////////////////////////////////////////////////////
// ******************   FUNZIONI INIZIALIZZAZIONE *******************
/////////////////////////////////////////////////////////////////////
  //funzione per trovare su quale porta  installata la scheda
  int detect_baseio(void)
  {
    static int val[] = {0x210, 0x220, 0x230, 0x240, 0x250, 0x260, 0x280};
    static int count = sizeof(val)/sizeof(*val) ;
    int i;

    for (i = 0; i < count; i++)
      if (reset_dsp(val[i])) {
	    return val[i];
	    break;}
	if (i==count) return 0;    	    
  }
  
//funzione di inizializzazione SB16
void trova_sb16(void)
{
	des_sb *p_des;
	bool r1,r2;
	p_des=&sb_16;
	int vett[2];
    	int baseio;
    	int detected;
	int versione;
	char versione_h, versione_l;
	unsigned char prova_irq,prova_dma,prova_dma_tmp;
			
	printk("Probing scheda audio SB16 compatibile...\n");
	baseio = 0;
	baseio = detect_baseio();
	if (baseio != 0) { //inizializzazione descrittore
		printk("Indirizzo di porta %x",baseio);
		p_des->indreg.FM_STS_1=(char*)baseio;
	        p_des->indreg.FM_DATA_1=(char*)(baseio+0x1);
		p_des->indreg.FM_AD_STS=(char*)(baseio+0x2);
		p_des->indreg.FM_AD_DATA=(char*)(baseio+0x3);
		p_des->indreg.MIX_REG=(char*)(baseio+0x4);
		p_des->indreg.MIX_DATA=(char*)(baseio+0x5);
		p_des->indreg.DSP_RESET=(char*)(baseio+0x6);
		p_des->indreg.RES_1=(char*)(baseio+0x7);
		p_des->indreg.FM_STS_2=(char*)(baseio+0x8);
		p_des->indreg.FM_DATA_2=(char*)(baseio+0x9);
		p_des->indreg.DSP_READ=(char*)(baseio+0xa);
		p_des->indreg.RES_2=(char*)(baseio+0xb);
		p_des->indreg.DSP_WRITE=(char*)(baseio+0xc);
		p_des->indreg.RES_3=(char*)(baseio+0xd); 
		p_des->indreg.DSP_READ_STS=(char*)(baseio+0xe);
		p_des->indreg.RES_4=(char*)(baseio+0xf);
		p_des->indreg.CD_DATA=(char*)(baseio+0x10);
		p_des->indreg.CD_STS=(char*)(baseio+0x11);
		p_des->indreg.CD_RESET=(char*)(baseio+0x12);
		p_des->indreg.CD_ENABLE=(char*)(baseio+0x13);
		prova_irq=getMixerSetting(0x80);  //Leggo la configurazione IRQ
		switch (prova_irq){
			case 0x01: 
				p_des->irq=2;
				break;
			case 0x02:
				p_des->irq=5;			
				break;				
			case 0x04:
				p_des->irq=7;
				break;				
			case 0x08:
				p_des->irq=10;			
				break;				
		}
		prova_dma=getMixerSetting(0x81); //Leggo la configurazione DMA		
		prova_dma_tmp=(prova_dma & 0x0F);
		switch (prova_dma_tmp){
			case 0x01: 
				p_des->dma=0;
				break;
			case 0x02:
				p_des->dma=1;
				break;				
			case 0x08:
				p_des->dma=3;
				break;				
			}
		prova_dma_tmp=(prova_dma & 0xF0)>>4;
		switch (prova_dma_tmp){
			case 0x00: 
				p_des->dma16=0;
				break;
			case 0x01:
				p_des->dma16=5;
				break;				
			case 0x02:
				p_des->dma16=6;
				break;				
			case 0x04:
				p_des->dma16=7;
				break;				
			}
		printk(" DMA %d/%d ",p_des->dma,p_des->dma16);
		printk("IRQ %d\n",p_des->irq);	}
	else {
		printk ("Non trovata SB16 compatibile\n");
		return;
		}	
	
	write_dsp (0xe1);	//ricerca versione DSP			
	versione_h = read_dsp ();
	versione_l = read_dsp ();
	printk("Versione DSP %d.%d\n",versione_h,versione_l);	

	p_des->indreg.MIDI_STS=(char*)0x331; //MIDI	
	p_des->indreg.MIDI_DATA=(char*)0x330;
	if (cmd_midi(MIDI_RESET)) {
		printk ("Trovata interfaccia MIDI MPU-401 su porta 0x330\n");
	}
	else {
		p_des->indreg.MIDI_STS=(char*)0x301;
		p_des->indreg.MIDI_DATA=(char*)0x300;
		if (cmd_midi(MIDI_RESET)) {
			printk ("Trovata interfaccia MIDI MPU-401 su porta 0x300\n");
		}
		else 
			printk ("Non trovata interfaccia MIDI MPU-401\n");		
	}
	

	sem_ini(p_des->mutex, 1, r1); 	//Inizializzazione semafori
	sem_ini(p_des->sincr, 0, r2);
	if(!r1 || !r2)
		panic("\nImpossibile allocare i semafori per l' IO");
	
	int SB_IRQ = p_des->irq; //attivazione processo esterno	
	short id;
	activate_pe(dmaest_sb,0,PRIO_ESTERN_BASE+IRQ_MAX-SB_IRQ,LIV_SISTEMA,id,sb,r1);
	if (!r1 || !r2)
		panic("\nImpossibile creare un processo esterno per SB16");
		
	p_des->dmareg.iMAR=(char*)dma_chan_ind[p_des->dma-1][0] ; //inizializzazione registri DMA
	p_des->dmareg.iTCR=(char*)dma_chan_ind[p_des->dma-1][1] ;
	p_des->dmareg.iPAGE=(char*)dma_chan_ind[p_des->dma-1][2] ;
	p_des->dmareg.iCLEAR_FF=(char*)dma_chan_ind[p_des->dma-1][3] ;			
	
	unsigned char irq1=0x1;
	irq1 = irq1 << (unsigned char)p_des->irq;
	irq1 = ~irq1;
	abilita_sb(irq1); //abilitazione a generare interruzioni
	
}

////////////////////////////////////////////////////////////////////////////////
//                 INIZIALIZZAZIONE DEL SOTTOSISTEMA DI I/O                   //
////////////////////////////////////////////////////////////////////////////////

// inizializza i gate usati per le chiamate di IO
//
extern "C" void fill_io_gates(void);

// eseguita in fase di inizializzazione ad interruzioni disabilitate
//
extern "C" void io_init(void)
{	
	fill_io_gates();
	term_init();
	com_init();
	hd_init();
	dma_init();
	fdc_init();
	trova_sb16();
}

// Replicata in sistema.cpp
//
void *memcpy(void *dest, const void *src, unsigned int n)
{
	char *dest_ptr = (char *)dest, *src_ptr = (char *)src;
	int i;

	for(i = 0; i < n; ++i)
		*dest_ptr++ = *src_ptr++;

	return dest;
}

// Replicata in sistema.cpp
//
void *memset(void *dest, int c, size_t n)
{
        size_t i;

        for(i = 0; i < n; ++i)
              ((char *)dest)[i] = (char)c;

        return dest;
}





  
  
  